import 'reflect-metadata';
import { RMQ_TRANSPORT_DTO_TYPE, CrudRequestTypes } from '../common/constants';
import { TargetObject, ValidationDto } from '../common/types';
import { CUSTOM_DTO_DECORATOR_IS_REQUIRED, NO_BASE_NO_CUSTOM_DTO } from '../common/util/messages';
import { validateMethodArguments } from '../validation/validate-method-arguments';
/**
 * Save metadata for each property of Descriptor
 *
 * @param descriptor - Descriptor of original method
 */
function savePropertyMetadata(descriptor: PropertyDescriptor): Record<string, unknown> {
  const metadataFieldValues: Record<string, unknown> = {};

  const fields = Reflect.getMetadataKeys(descriptor.value) || [];
  for (let i = 0; i < fields.length; i += 1) {
    metadataFieldValues[fields[i]] = Reflect.getMetadata(fields[i], descriptor.value);
  }

  return metadataFieldValues;
}

/**
 * Assign metadata of Original Method to new Descriptor
 *
 * @param descriptor - Descriptor of original method
 * @param metadataFieldValues - Object with metadata fields list of original method
 */
function assignOldMetadataToNewDescriptor(
  descriptor: PropertyDescriptor,
  metadataFieldValues: Record<string, unknown> = {},
): void {
  const keys: string[] = Object.keys(metadataFieldValues);

  for (let i = 0; i < keys.length; i += 1) {
    Reflect.defineMetadata(keys[i], metadataFieldValues[keys[i]], descriptor.value);
  }
}

/**
 * Resolve method group by method name
 *
 * @param methodName - Method name
 */
function detectMethodGroup(methodName: string): CrudRequestTypes | undefined {
  switch (methodName) {
    case 'getMany':
    case 'getOne':
      return CrudRequestTypes.READ;
    case 'createOne':
    case 'createMany':
      return CrudRequestTypes.CREATE;
    case 'updateOne':
    case 'replaceOne':
      return CrudRequestTypes.UPDATE;
    case 'deleteOne':
      return CrudRequestTypes.DELETE;
    default:
      return undefined;
  }
}

/**
 * Validate custom dto requirement
 * @param target - target object
 * @param propertyName - property name
 * @param BaseDto - base validation dto
 */
function checkIFCustomDtoIsAvailable(
  target: TargetObject,
  propertyName: string,
  BaseDto?: ValidationDto,
): void {
  const customDtoClass = Reflect.getMetadata(
    RMQ_TRANSPORT_DTO_TYPE,
    target.prototype,
    propertyName,
  );
  const methodGroup: CrudRequestTypes | undefined = detectMethodGroup(propertyName);

  if (BaseDto && methodGroup && !customDtoClass) return;

  if (!customDtoClass && !methodGroup) {
    throw new Error(CUSTOM_DTO_DECORATOR_IS_REQUIRED(propertyName));
  }

  if (!BaseDto && !customDtoClass) {
    throw new Error(NO_BASE_NO_CUSTOM_DTO(propertyName));
  }
}

/**
 * Base decorator that should be added to controller to use library flow
 * @param validationDto - Validation DTO class
 */
export const MicroserviceValidation = (validationDto?: ValidationDto) => (target: TargetObject) => {
  const propertiesKeys = Object.getOwnPropertyNames(target.prototype);
  for (let i = 0; i < propertiesKeys.length; i += 1) {
    const descriptor: PropertyDescriptor | undefined = Object.getOwnPropertyDescriptor(
      target.prototype,
      propertiesKeys[i],
    );
    if (
      descriptor &&
      descriptor.value instanceof Function &&
      descriptor.value !== target.prototype.constructor
    ) {
      const metadataFieldValues = savePropertyMetadata(descriptor);
      const originalMethod = descriptor.value;

      descriptor.value = async function value(...args: any[]) {
        const customDtoClass = Reflect.getMetadata(
          RMQ_TRANSPORT_DTO_TYPE,
          target.prototype,
          propertiesKeys[i],
        );
        const methodGroup: CrudRequestTypes | undefined = detectMethodGroup(propertiesKeys[i]);
        await validateMethodArguments(args, {
          validationDto,
          methodGroup,
          customDtoClass,
          methodName: propertiesKeys[i],
        });
        return originalMethod.apply(this, args);
      };

      assignOldMetadataToNewDescriptor(descriptor, metadataFieldValues);

      Object.defineProperty(target.prototype, propertiesKeys[i], descriptor);

      // Check if custom dto is added and valid
      checkIFCustomDtoIsAvailable(target, propertiesKeys[i], validationDto);
    }
  }
};
