import { getMetadataStorage, MetadataStorage } from 'class-validator';
import { DtoInstance, ValidationDto, ValidationMetaDataCollector } from '../common/types';
import { ValidationMetadata } from '../interfaces/validation-metadata.interface';
import { CrudRequestTypes, VALIDATE_IF_RELATED_ONLY } from '../common/constants';

/**
 *
 * Get dto instance metadata
 *
 * @param dtoInstance - instance of dto class
 * @param metadataStorage - validation metadata storage
 */
function receiveTargetMetadata(
  dtoInstance: DtoInstance,
  metadataStorage: MetadataStorage,
): ValidationMetadata[] {
  const targetMetadata = [];
  let instancePrototype = Object.getPrototypeOf(dtoInstance).constructor;
  while (
    Object.prototype.hasOwnProperty.call(instancePrototype, 'name') &&
    instancePrototype.name.length
  ) {
    targetMetadata.push(
      ...metadataStorage.getTargetValidationMetadatas(instancePrototype, '', false, false),
    );
    if (
      !Object.getPrototypeOf(instancePrototype).name ||
      Object.getPrototypeOf(instancePrototype).name === 'Function' ||
      Object.getPrototypeOf(instancePrototype).name === instancePrototype.name
    ) {
      break;
    }
    instancePrototype = Object.getPrototypeOf(instancePrototype);
  }
  return targetMetadata;
}

/**
 * Receive properties keys of DTO
 *
 * @param dtoInstance - instance of DTO used to validate arguments
 * @param DtoForValidation -  Validation dto
 * @param group? - CRUD Request method type
 */
export function getDtoPropertiesFromInstance(
  dtoInstance: Record<string | symbol, any>,
  DtoForValidation: ValidationDto,
  group?: CrudRequestTypes,
): ValidationMetaDataCollector {
  const metadataStorage: MetadataStorage = getMetadataStorage();
  const targetMetadata = receiveTargetMetadata(dtoInstance, metadataStorage);
  const groupedMetadata: ValidationMetaDataCollector =
    metadataStorage.groupByPropertyName(targetMetadata);

  if (!group) return groupedMetadata;

  const properties = Object.getOwnPropertyNames(groupedMetadata);
  const metadataKeys: ValidationMetaDataCollector = {};

  for (let i = 0; i < properties.length; i += 1) {
    let isMainButValidateIfNestedOnly = false;

    if (dtoInstance.$isNestedDto === false) {
      const nestedMetadata = Reflect.getMetadata(
        VALIDATE_IF_RELATED_ONLY,
        DtoForValidation.prototype,
        properties[i],
      );

      if (
        nestedMetadata?.groups &&
        nestedMetadata.groups.length &&
        nestedMetadata.groups.includes(group)
      ) {
        isMainButValidateIfNestedOnly = true;
      }
    }

    if (
      groupedMetadata[properties[i]].some(
        (item) =>
          item.always ||
          !item.groups?.length ||
          (item.groups.indexOf(group) !== -1 && !isMainButValidateIfNestedOnly),
      )
    ) {
      metadataKeys[properties[i]] = groupedMetadata[properties[i]];
    }
  }

  return group ? metadataKeys : groupedMetadata;
}
