import { ValidationTypes } from 'class-validator';
import {
  DtoInstance,
  TargetObject,
  ValidationDto,
  ValidationMetaDataCollector,
} from '../../common/types';
import { CreateDtoInstanceOptions } from '../../interfaces/create-dto-instance-options.interface';
import { RelatedDtoOptions } from '../../interfaces/related-dto-options.interface';
import { RELATED_DTO_OPTIONS } from '../../common/constants';
import { generatePrefix, generateWhitelistError } from '../validation-errors';
import { getDtoPropertiesFromInstance } from '../get-dto-properties-from-instance';
import { getNestedRelations } from '../../common/util/fetch-property-relations';

/**
 * Create dto instance from object
 * @param target - targetObject
 * @param DtoForValidation = Dto class for validation
 * @param options - Options object for dto creating
 * @param isInnerRun - Specify if method is running inside validator flow
 */
export function createDtoInstance(
  target: TargetObject,
  DtoForValidation: ValidationDto,
  options: CreateDtoInstanceOptions,
  isInnerRun = false,
): DtoInstance {
  const dtoInstance = new DtoForValidation();
  if (isInnerRun) {
    dtoInstance.$isNestedDto = !options.isMainDto;
  }
  const metaKeysValidationData: ValidationMetaDataCollector = getDtoPropertiesFromInstance(
    dtoInstance,
    DtoForValidation,
    options.methodGroup,
  );
  const metaKeysList: string[] = Object.getOwnPropertyNames(metaKeysValidationData);
  const targetKeys: string[] = Object.keys(target);
  for (let i = 0; i < targetKeys.length; i += 1) {
    if (
      Object.prototype.hasOwnProperty.call(target, targetKeys[i]) &&
      metaKeysList.includes(targetKeys[i])
    ) {
      const nestedValidationParam = metaKeysValidationData[targetKeys[i]].find(
        (item: { type: string }) => item.type === ValidationTypes.NESTED_VALIDATION,
      );
      if (nestedValidationParam && nestedValidationParam.context) {
        // Check for relation metadata
        const relatedDtoParamOptions: RelatedDtoOptions | undefined = Reflect.getMetadata(
          RELATED_DTO_OPTIONS,
          DtoForValidation.prototype,
          targetKeys[i],
        );

        // Prepare CreateDtoInstanceOptions for nested object
        const nestedValidationOptions: CreateDtoInstanceOptions = {
          validationErrors: options.validationErrors,
          isCustomDto: options.isCustomDto,
          methodGroup: options.methodGroup,
          nestedRelations: options.isMainDto
            ? relatedDtoParamOptions?.nestedRelations || []
            : getNestedRelations(targetKeys[i], options.nestedRelations),
          parentPropertyPrefix: generatePrefix(targetKeys[i], options.parentPropertyPrefix),
        };

        if (
          relatedDtoParamOptions &&
          !options.isMainDto &&
          !options.nestedRelations?.includes(targetKeys[i])
        ) {
          // Generate error in case if relation metadata exists but target object should not contain any relations
          options.validationErrors.push(
            generateWhitelistError(
              target,
              targetKeys[i],
              target[targetKeys[i]],
              options.parentPropertyPrefix,
            ),
          );
        } else if (
          nestedValidationParam.each &&
          Array.isArray(target[targetKeys[i]]) &&
          target[targetKeys[i]].length
        ) {
          // Create dto for each property in case if each array value should be validated
          const properties: TargetObject = target[targetKeys[i]];
          const propertiesKeys = Object.keys(properties);
          const propertyInstances: DtoInstance[] = [];

          for (let j = 0; j < propertiesKeys.length; j += 1) {
            propertyInstances.push(
              createDtoInstance(
                properties[j],
                nestedValidationParam.context,
                {
                  ...nestedValidationOptions,
                  parentPropertyPrefix: generatePrefix(
                    targetKeys[i],
                    options.parentPropertyPrefix,
                    j,
                  ),
                },
                isInnerRun,
              ),
            );
          }
          dtoInstance[targetKeys[i]] = propertyInstances;
        } else if (
          typeof target[targetKeys[i]] === 'object' &&
          target[targetKeys[i]] != null &&
          !Array.isArray(target[targetKeys[i]])
        ) {
          // Create dto for nested property in case of single object
          dtoInstance[targetKeys[i]] = createDtoInstance(
            target[targetKeys[i]],
            nestedValidationParam.context,
            nestedValidationOptions,
            isInnerRun,
          );
        } else {
          dtoInstance[targetKeys[i]] = target[targetKeys[i]];
        }
      } else {
        dtoInstance[targetKeys[i]] = target[targetKeys[i]];
      }
    } else {
      options.validationErrors.push(
        generateWhitelistError(
          target,
          targetKeys[i],
          target[targetKeys[i]],
          options.parentPropertyPrefix,
        ),
      );
    }
  }

  return dtoInstance;
}
