import { validate, ValidationError } from 'class-validator';
import { DtoInstance, TargetObject, ValidationDto } from '../../common/types';
import { ValidateArgumentOptions } from '../../interfaces/validate-argument-options.interface';
import { createDtoInstance } from './create-dto-instance';

/**
 * Add prefix to error message (for bulk validation)
 * @param validationError - Error object
 * @param prefix - prefix string
 */
function addPrefixToErrorMessage(
  validationError: ValidationError,
  prefix: string,
): ValidationError {
  if (validationError.constraints) {
    const constraintsKeys = Object.keys(validationError.constraints);
    for (let i = 0; i < constraintsKeys.length; i += 1) {
      validationError.constraints[constraintsKeys[i]] =
        prefix + validationError.constraints[constraintsKeys[i]];
    }
  }
  return validationError;
}

/**
 * Function for dto validation
 * @param dto - dto instance for validation
 * @param validationErrors - Base array of validation errors
 * @param options - Validate argument options object
 */
export async function validateDto(
  dto: TargetObject,
  validationErrors: ValidationError[],
  options: ValidateArgumentOptions,
): Promise<void> {
  const isCustomDto = !!options.customDtoClass;
  const DtoForValidation = isCustomDto ? options.customDtoClass : options.validationDto;

  const dtoInstance: DtoInstance = createDtoInstance(
    dto,
    DtoForValidation as ValidationDto,
    {
      validationErrors,
      methodGroup: options.methodGroup,
      isCustomDto,
      isMainDto: true,
    },
    true,
  );

  const dtoValidationErrors: ValidationError[] = await validate(dtoInstance, {
    groups: options.methodGroup ? [options.methodGroup] : [],
    forbidUnknownValues: false,
  });

  if (options.errorPrefix) {
    validationErrors.push(
      ...dtoValidationErrors.map((error) =>
        addPrefixToErrorMessage(error, options.errorPrefix as string),
      ),
    );
  } else {
    validationErrors.push(...dtoValidationErrors);
  }
}
