import { ValidationError } from 'class-validator';
import { generatePrefix, parseValidationErrors } from './validation-errors';
import { ValidateArgumentOptions } from '../interfaces/validate-argument-options.interface';
import { validateCrudRequest } from './validate-request/validate-crud-request';
import { validateDto } from './validate-dto/validate-dto';

/**
 * Check if property is request object
 * @param property - Object for checking
 */
function isPropertyRequestObject(property: any): boolean {
  return (
    Object.prototype.hasOwnProperty.call(property, 'parsed') &&
    Object.prototype.hasOwnProperty.call(property, 'options')
  );
}
/**
 * Validation of original method arguments
 *
 * @param argumentsList - list of method arguments
 * @param options - Validate argument options
 */
export async function validateMethodArguments(
  argumentsList: any[],
  options: ValidateArgumentOptions,
): Promise<void> {
  const validationErrors: ValidationError[] = [];
  const validationPromises: Promise<void>[] = [];

  for (let i = 0; i < argumentsList.length; i += 1) {
    const argument = argumentsList[i];
    if (isPropertyRequestObject(argument)) {
      validationPromises.push(validateCrudRequest(argument.parsed, validationErrors, options));
    }
    if (
      Object.prototype.hasOwnProperty.call(argument, 'req') &&
      isPropertyRequestObject(argument.req)
    ) {
      validationPromises.push(validateCrudRequest(argument.req.parsed, validationErrors, options));
    }
    if (Object.prototype.hasOwnProperty.call(argument, 'dto') && argument.dto) {
      if (
        Object.prototype.hasOwnProperty.call(argument.dto, 'bulk') &&
        Array.isArray(argument.dto.bulk) &&
        options.methodName === 'createMany'
      ) {
        for (let j = 0; j < argument.dto.bulk.length; j += 1) {
          const dto = argument.dto.bulk[i];
          validationPromises.push(
            validateDto(dto, validationErrors, {
              ...options,
              errorPrefix: generatePrefix('bulk', undefined, j),
            }),
          );
        }
      } else {
        validationPromises.push(validateDto(argument.dto, validationErrors, options));
      }
    }
  }

  await Promise.all(validationPromises);

  if (validationErrors && validationErrors.length) {
    parseValidationErrors(validationErrors);
  }
}
