import { ValidationError } from 'class-validator';
import { ValidateArgumentOptions } from '../../interfaces/validate-argument-options.interface';
import {
  ParsedRequestParams,
  QueryFilter,
  SCondition,
} from '../../interfaces/parsed-request-params.interface';
import { QueryFilterOperators, RequestFieldTypes } from '../../common/constants';
import { validateRequestParam } from './validate-request-param';
import { ConvertedRequestParam, TargetObject, ValidationDto } from '../../common/types';
import { validateFieldsExisting } from './vilidate-fields-existing';
import { generateNotPrimitiveParamError } from '../validation-errors';

/**
 * Fetch and convert queries from filters arrays
 * @param requestFilters - Filters array
 * @param convertedParamsArray - Array for converted queries
 */
function convertFilters(
  requestFilters: QueryFilter[],
  convertedParamsArray: ConvertedRequestParam[],
): void {
  for (let i = 0; i < requestFilters.length; i += 1) {
    const item: QueryFilter = requestFilters[i];
    switch (item.operator) {
      case QueryFilterOperators.BETWEEN:
      case QueryFilterOperators.IN:
      case QueryFilterOperators.IN_LOWER:
      case QueryFilterOperators.NOT_IN:
      case QueryFilterOperators.NOT_IN_LOWER:
        for (let k = 0; k < item.value.length; k += 1) {
          convertedParamsArray.push({ [item.field]: item.value[k] });
        }
        break;
      case QueryFilterOperators.IS_NULL:
      case QueryFilterOperators.NOT_NULL:
        convertedParamsArray.push({ [item.field]: null });
        break;
      default:
        convertedParamsArray.push({ [item.field]: item.value });
        break;
    }
  }
}

/**
 * Validate if query value contains primitive value
 * @param requestParam - converted request parameter
 */
function validatePrimitiveValues(requestParam: ConvertedRequestParam): ValidationError[] {
  const validationErrors = [];

  const keys = Object.keys(requestParam);
  for (let i = 0; i < keys.length; i += 1) {
    const value = requestParam[keys[i]];
    if (typeof value === 'object' && value !== null) {
      validationErrors.push(generateNotPrimitiveParamError(requestParam, keys[i], value, ''));
    }
  }

  return validationErrors;
}

/**
 * Parse of Search Params Operators Recursively
 *
 * @param target - property with search operators and conditions
 * @param convertedRequestParams - Array of converted request parameters
 * @param propertyName - property name
 */
function recursiveSearchPropertiesParse(
  target: any,
  convertedRequestParams: ConvertedRequestParam[],
  propertyName: string,
): void {
  const keys = Object.keys(target);
  for (let i = 0; i < keys.length; i += 1) {
    const value = target[keys[i]];
    if (!Array.isArray(value) && typeof value === 'object' && value !== null) {
      recursiveSearchPropertiesParse(value, convertedRequestParams, propertyName);
    } else if (keys[i] === '$isnull' || keys[i] === '$notnull') {
      convertedRequestParams.push({ [propertyName]: null });
    } else if (Array.isArray(value) && value !== null) {
      convertedRequestParams.push(...value.map((item) => ({ [propertyName]: item })));
    } else {
      convertedRequestParams.push({ [propertyName]: value });
    }
  }
}

/**
 * Parse of Search Params Recursively
 *
 * @param properties - object containing search param properties
 * @param convertedRequestParams - Array of converted request parameters
 */
function recursiveSearchParse(
  properties: SCondition[],
  convertedRequestParams: ConvertedRequestParam[],
) {
  for (let i = 0; i < properties.length; i += 1) {
    const property: SCondition = properties[i];
    const objectKeys: string[] = Object.keys(property);

    for (let k = 0; k < objectKeys.length; k += 1) {
      const key = objectKeys[k];
      if (key !== '$and' && key !== '$or') {
        if (typeof property === 'object' && property !== null) {
          recursiveSearchPropertiesParse(property, convertedRequestParams, key);
        }
      } else {
        recursiveSearchParse(property[key] as SCondition[], convertedRequestParams);
      }
    }
  }
}

/**
 * Parse CRUD Request Search param
 *
 * @param search - CRUD Request Search param
 * @param convertedRequestParams - Array of converted request parameters
 */
function convertSearch(search: SCondition, convertedRequestParams: ConvertedRequestParam[]): void {
  if (Object.prototype.hasOwnProperty.call(search, '$and') && Array.isArray(search.$and)) {
    const properties: SCondition[] = search.$and.filter(
      (item: any) => item && JSON.stringify(item) !== '{}',
    );
    recursiveSearchParse(properties, convertedRequestParams);
  }
}

/**
 * Validate crud request parsed properties
 * @param parsedRequest - parsed request object
 * @param validationErrors - validation errors array
 * @param options - validation options object
 */
export async function validateCrudRequest(
  parsedRequest: ParsedRequestParams,
  validationErrors: ValidationError[],
  options: ValidateArgumentOptions,
): Promise<void> {
  const isCustomDto = !!options.customDtoClass;
  const DtoForValidation = isCustomDto
    ? (options.customDtoClass as ValidationDto)
    : (options.validationDto as ValidationDto);

  const convertedRequestParams: ConvertedRequestParam[] = [];

  convertFilters(
    [...parsedRequest.filter, ...parsedRequest.paramsFilter, ...parsedRequest.or],
    convertedRequestParams,
  );

  convertSearch(parsedRequest.search, convertedRequestParams);

  const validationOptions = {
    validationErrors,
    methodGroup: options.methodGroup,
    isCustomDto,
    isMainDto: true,
  };

  // Validate request fields & sort
  const fieldsObject: TargetObject = {};
  parsedRequest.fields.forEach((f) => {
    fieldsObject[f] = RequestFieldTypes.FIELD;
  });
  parsedRequest.sort.forEach((s) => {
    fieldsObject[s.field] = RequestFieldTypes.SORT;
  });
  // Joins should be added last. Join properties should not be rewritten by sort and fields properties.
  // However sort and fields properties could be rewritten by join properties.
  parsedRequest.join.forEach((j) => {
    fieldsObject[j.field] = RequestFieldTypes.JOIN;
  });
  validateFieldsExisting(fieldsObject, DtoForValidation, validationOptions);

  // Validate each of converted parameters
  const validationPromises: Promise<void>[] = [];

  for (let i = 0; i < convertedRequestParams.length; i += 1) {
    const primitiveValueErrors: ValidationError[] = validatePrimitiveValues(
      convertedRequestParams[i],
    );
    if (primitiveValueErrors.length) {
      validationErrors.push(...primitiveValueErrors);
    } else {
      validationPromises.push(
        validateRequestParam(convertedRequestParams[i], DtoForValidation, validationOptions),
      );
    }
  }

  await Promise.all(validationPromises);
}
