import {
  getMetadataStorage,
  MetadataStorage,
  ValidationArguments,
  ValidationTypes,
} from 'class-validator';
import {
  ConvertedRequestParam,
  ValidationDto,
  ValidationMetaDataCollector,
} from '../../common/types';
import { getDtoPropertiesFromInstance } from '../get-dto-properties-from-instance';
import { ValidationMetadata } from '../../interfaces/validation-metadata.interface';
import {
  generateConstraintError,
  generatePrefix,
  generateWhitelistError,
} from '../validation-errors';
import { CreateDtoInstanceOptions } from '../../interfaces/create-dto-instance-options.interface';
import { RelatedDtoOptions } from '../../interfaces/related-dto-options.interface';
import { RELATED_DTO_OPTIONS } from '../../common/constants';
import { replaceDotNotationField } from '../../common/util/replace-dot-notation-field';
import { getNestedRelations } from '../../common/util/fetch-property-relations';

/**
 * Replaces special tokens in default error message with real values
 *
 * @param message - default text error message with special tokens
 * @param validationArguments - arguments used for validation
 */
function replaceMessageSpecialTokens(
  message: string | ((args: ValidationArguments) => string),
  validationArguments: ValidationArguments,
): string {
  let messageString = '';
  if (message instanceof Function) {
    messageString = (message as (args: ValidationArguments) => string)(validationArguments);
  } else if (typeof message === 'string') {
    messageString = message;
  }

  if (messageString && Array.isArray(validationArguments.constraints)) {
    validationArguments.constraints.forEach((constraint, index) => {
      messageString = messageString.replace(
        new RegExp(`\\$constraint${index + 1}`, 'g'),
        constraint,
      );
    });
  }
  if (messageString) {
    messageString = messageString
      .replace(/\$value/g, JSON.stringify(validationArguments.value))
      .replace(/\$property/g, validationArguments.property)
      .replace(/\$target/g, validationArguments.targetName);
  }

  return messageString;
}

/**
 * Receives constraint type by metadata info
 *
 * @param metadata - validation metadata of property
 * @param customValidatorMetadata - constraint metadata
 */
function getConstraintType(metadata: ValidationMetadata, customValidatorMetadata?: any): string {
  return customValidatorMetadata && customValidatorMetadata.name
    ? customValidatorMetadata.name
    : metadata.type;
}

/**
 * Check if property should be validated
 * @param metadataArray - Array of property metadata
 * @param requestParam - Property parent object
 * @param options - Options object
 */
function shouldPropertyBeValidated(
  metadataArray: ValidationMetadata[],
  requestParam: ConvertedRequestParam,
  options: CreateDtoInstanceOptions,
): boolean {
  const filteredMetaKeys = metadataArray.filter(
    (item: ValidationMetadata) =>
      item.type === ValidationTypes.CONDITIONAL_VALIDATION &&
      Array.isArray(item.constraints) &&
      item.constraints[0] instanceof Function &&
      (options.isCustomDto || item.groups?.includes(options.methodGroup as string)),
  );

  let shouldBeValidated = true;

  for (let j = 0; j < filteredMetaKeys.length; j += 1) {
    const metaData = filteredMetaKeys[j];
    if (!metaData.constraints[0](requestParam, requestParam[metaData.propertyName])) {
      shouldBeValidated = false;
    }
  }

  return shouldBeValidated;
}

/**
 * Validate single request parameter
 * @param requestParam - request parameter object
 * @param DtoForValidation - Dto class for validation
 * @param options - create dto instance options object
 */
export async function validateRequestParam(
  requestParam: ConvertedRequestParam,
  DtoForValidation: ValidationDto,
  options: CreateDtoInstanceOptions,
) {
  const dtoInstance = new DtoForValidation();
  dtoInstance.$isNestedDto = !options.isMainDto;
  const nestedValidationPromises: Promise<void>[] = [];

  const metaDataStorage: MetadataStorage = getMetadataStorage();
  const metaKeysValidationData: ValidationMetaDataCollector = getDtoPropertiesFromInstance(
    dtoInstance,
    DtoForValidation,
    options.methodGroup,
  );

  const targetKeys: string[] = Object.keys(requestParam);
  for (let i = 0; i < targetKeys.length; i += 1) {
    // In case if field contains '.' its value is replaced by object instead and key itself is replaced by new one.
    const key =
      targetKeys[i] != null && targetKeys[i].includes('.')
        ? replaceDotNotationField(targetKeys[i], requestParam)
        : targetKeys[i];

    if (
      Object.prototype.hasOwnProperty.call(metaKeysValidationData, key) &&
      metaKeysValidationData[key] &&
      metaKeysValidationData[key].length
    ) {
      if (shouldPropertyBeValidated(metaKeysValidationData[key], requestParam, options)) {
        const filteredMetaKeys = metaKeysValidationData[key].filter(
          (item: ValidationMetadata) =>
            item.type === ValidationTypes.NESTED_VALIDATION ||
            (!!item.constraintCls &&
              (options.isCustomDto || item.groups?.includes(options.methodGroup as string))),
        );

        for (let j = 0; j < filteredMetaKeys.length; j += 1) {
          const metaData: ValidationMetadata = filteredMetaKeys[j];
          const constraints = metaDataStorage.getTargetValidatorConstraints(metaData.constraintCls);
          if (metaData.type === ValidationTypes.NESTED_VALIDATION && metaData.context) {
            const relatedDtoParamOptions: RelatedDtoOptions | undefined = Reflect.getMetadata(
              RELATED_DTO_OPTIONS,
              DtoForValidation.prototype,
              key,
            );

            if (
              relatedDtoParamOptions &&
              !options.isMainDto &&
              !options.nestedRelations?.includes(key)
            ) {
              options.validationErrors.push(
                generateWhitelistError(
                  requestParam,
                  key,
                  requestParam[key],
                  options.parentPropertyPrefix,
                ),
              );
            } else if (typeof requestParam[key] === 'object' && requestParam[key] !== null) {
              nestedValidationPromises.push(
                validateRequestParam(requestParam[key], metaData.context, {
                  validationErrors: options.validationErrors,
                  isCustomDto: options.isCustomDto,
                  methodGroup: options.methodGroup,
                  nestedRelations: options.isMainDto
                    ? relatedDtoParamOptions?.nestedRelations || []
                    : getNestedRelations(key, options.nestedRelations),
                  parentPropertyPrefix: generatePrefix(key, options.parentPropertyPrefix),
                }),
              );
            }
          }
          for (let k = 0; k < constraints.length; k += 1) {
            const constraint = constraints[i];
            const validationArguments: ValidationArguments = {
              targetName: DtoForValidation.name,
              property: metaData.propertyName,
              object: dtoInstance,
              value: requestParam[key],
              constraints: metaData.constraints,
            };
            const result = constraint.instance.validate(requestParam[key], validationArguments);
            if (
              !result &&
              constraint &&
              constraint.instance &&
              constraint.instance.defaultMessage
            ) {
              const message: string | undefined =
                constraint.instance.defaultMessage(validationArguments);
              options.validationErrors.push(
                generateConstraintError(
                  requestParam,
                  requestParam[key],
                  {
                    propertyName: key,
                    type: getConstraintType(metaData, constraint),
                    message: replaceMessageSpecialTokens(message, validationArguments),
                  },
                  options.parentPropertyPrefix,
                ),
              );
            }
          }
        }
      }
    } else {
      options.validationErrors.push(
        generateWhitelistError(requestParam, key, requestParam[key], options.parentPropertyPrefix),
      );
    }
  }

  await Promise.all(nestedValidationPromises);
}
