import { ValidationTypes } from 'class-validator';
import { TargetObject, ValidationDto, ValidationMetaDataCollector } from '../../common/types';
import { getDtoPropertiesFromInstance } from '../get-dto-properties-from-instance';
import { CreateDtoInstanceOptions } from '../../interfaces/create-dto-instance-options.interface';
import { replaceDotNotationField } from '../../common/util/replace-dot-notation-field';
import {
  generatePrefix,
  generateWhitelistError,
  generateWrongJoinError,
} from '../validation-errors';
import { ValidationMetadata } from '../../interfaces/validation-metadata.interface';
import { RelatedDtoOptions } from '../../interfaces/related-dto-options.interface';
import { RELATED_DTO_OPTIONS, RequestFieldTypes } from '../../common/constants';
import { getNestedRelations } from '../../common/util/fetch-property-relations';

/**
 * Validate if property exists in dto
 * @param targetObject - target object
 * @param DtoForValidation - Dto class for validation
 * @param options - Create dto instance options
 */
export const validateFieldsExisting = (
  targetObject: TargetObject,
  DtoForValidation: ValidationDto,
  options: CreateDtoInstanceOptions,
): void => {
  const dtoInstance = new DtoForValidation();
  dtoInstance.$isNestedDto = !options.isMainDto;
  const metaKeysValidationData: ValidationMetaDataCollector = getDtoPropertiesFromInstance(
    dtoInstance,
    DtoForValidation,
    options.methodGroup,
  );

  const targetKeys: string[] = Object.keys(targetObject);
  for (let i = 0; i < targetKeys.length; i += 1) {
    // In case if field contains '.' its value is replaced by object instead and key itself is replaced by new one.
    const key =
      targetKeys[i] != null && targetKeys[i].includes('.')
        ? replaceDotNotationField(targetKeys[i], targetObject)
        : targetKeys[i];

    if (
      Object.prototype.hasOwnProperty.call(metaKeysValidationData, key) &&
      metaKeysValidationData[key] &&
      metaKeysValidationData[key].length
    ) {
      const filteredMetaKeys = metaKeysValidationData[key].filter(
        (item: ValidationMetadata) =>
          item.type === ValidationTypes.NESTED_VALIDATION ||
          (!!item.constraintCls &&
            (options.isCustomDto || item.groups?.includes(options.methodGroup as string))),
      );

      for (let j = 0; j < filteredMetaKeys.length; j += 1) {
        const metaData: ValidationMetadata = filteredMetaKeys[j];

        if (metaData.type === ValidationTypes.NESTED_VALIDATION && metaData.context) {
          const relatedDtoParamOptions: RelatedDtoOptions | undefined = Reflect.getMetadata(
            RELATED_DTO_OPTIONS,
            DtoForValidation.prototype,
            key,
          );
          if (
            relatedDtoParamOptions &&
            !options.isMainDto &&
            !options.nestedRelations?.includes(key)
          ) {
            options.validationErrors.push(
              generateWhitelistError(
                targetObject,
                key,
                targetObject[key],
                options.parentPropertyPrefix,
              ),
            );
          } else if (typeof targetObject[key] === 'object' && targetObject[key] !== null) {
            validateFieldsExisting(targetObject[key], metaData.context, {
              validationErrors: options.validationErrors,
              isCustomDto: options.isCustomDto,
              methodGroup: options.methodGroup,
              nestedRelations: options.isMainDto
                ? relatedDtoParamOptions?.nestedRelations || []
                : getNestedRelations(key, options.nestedRelations),
              parentPropertyPrefix: generatePrefix(key, options.parentPropertyPrefix),
            });
          }
        } else if (
          metaData.type !== ValidationTypes.NESTED_VALIDATION &&
          targetObject[key] === RequestFieldTypes.JOIN
        ) {
          options.validationErrors.push(
            generateWrongJoinError(
              targetObject,
              key,
              targetObject[key],
              options.parentPropertyPrefix,
            ),
          );
        }
      }
    } else {
      options.validationErrors.push(
        generateWhitelistError(targetObject, key, targetObject[key], options.parentPropertyPrefix),
      );
    }
  }
};
