import { ValidationError, ValidationTypes } from 'class-validator';
import { ConvertedRequestParam, TargetObject } from '../common/types';
import { GenerateConstraintErrorOptions } from '../interfaces/generate-constraint-error-options.interface';

type ConstraintsGeneratorFunc = (
  property: string,
  parentPrefix: string | undefined,
) => Record<string, string>;

type ValidationErrorGenerationFunc = (
  target: TargetObject,
  propertyName: string,
  value: any,
  parentPrefix: string | undefined,
) => ValidationError;

export class NestJSMicroserviceValidationException extends Error {
  constructor(public readonly messages: string[]) {
    super(messages.join(', '));
  }
}

/**
 * Generate prefix for nested property errors
 *
 * @param property - Property name
 * @param parentPrefix - Property prefix
 * @param arrayIndex - index of item in array
 */
export const generatePrefix = (
  property: string,
  parentPrefix: string | undefined = '',
  arrayIndex: number | undefined = undefined,
): string => `${parentPrefix}${property} -> ${arrayIndex !== undefined ? `${arrayIndex} -> ` : ''}`;

/**
 * Manually create validation error
 *
 * @param object - target object
 * @param value - value of property
 * @param options - error options (message,type,property)
 * @param parentPrefix - parent property prefix for error messages
 */
export function generateConstraintError(
  object: ConvertedRequestParam,
  value: any,
  options: GenerateConstraintErrorOptions,
  parentPrefix = '',
): ValidationError {
  const validationError = new ValidationError();
  validationError.property = options.propertyName;
  validationError.children = [];
  validationError.constraints = {};
  validationError.target = object;
  validationError.value = value;
  validationError.constraints[options.type] = parentPrefix + options.message;
  return validationError;
}

/**
 * Generate validation error factory function
 * @param constrainGenerator - Constraint generation function
 */
function generateValidationErrorFactory(
  constrainGenerator: ConstraintsGeneratorFunc,
): ValidationErrorGenerationFunc {
  return (
    target: TargetObject,
    propertyName: string,
    value: any,
    parentPrefix = '',
  ): ValidationError => {
    const validationError: ValidationError = new ValidationError();
    validationError.target = target;
    validationError.value = value;
    validationError.property = propertyName;
    validationError.constraints = constrainGenerator(propertyName, parentPrefix);
    validationError.children = [];
    return validationError;
  };
}
/**
 * Generate whiteList ValidationError
 *
 * @param target - DTO used for validation
 * @param propertyName - name of property
 * @param value - value of property
 * @param parentPrefix - parent property prefix for error messages
 */
export const generateWhitelistError: ValidationErrorGenerationFunc = generateValidationErrorFactory(
  (propertyName: string, parentPrefix = ''): Record<string, string> => ({
    [ValidationTypes.WHITELIST]: `${parentPrefix}property ${propertyName} should not exist`,
  }),
);

/**
 * Generate wrong join ValidationError
 *
 * @param target - DTO used for validation
 * @param propertyName - name of property
 * @param value - value of property
 * @param parentPrefix - parent property prefix for error messages
 */
export const generateWrongJoinError: ValidationErrorGenerationFunc = generateValidationErrorFactory(
  (propertyName: string, parentPrefix = ''): Record<string, string> => ({
    [ValidationTypes.WHITELIST]: `${parentPrefix}property ${propertyName} is not allowed for join`,
  }),
);

/**
 * Generate wrong join ValidationError
 *
 * @param target - DTO used for validation
 * @param propertyName - name of property
 * @param value - value of property
 * @param parentPrefix - parent property prefix for error messages
 */
export const generateNotPrimitiveParamError: ValidationErrorGenerationFunc =
  generateValidationErrorFactory(
    (propertyName: string, parentPrefix = ''): Record<string, string> => ({
      [ValidationTypes.CUSTOM_VALIDATION]: `${parentPrefix}value for property ${propertyName} should be primitive`,
    }),
  );

/**
 * Parse single ValidationError constraints to string
 *
 * @param parsedErrors - list of previously parsed errors
 * @param validationError - single validation error object
 * @param parentPrefix - parent prefix for error string
 */
export const parseValidationError = (
  parsedErrors: string[],
  validationError: ValidationError,
  parentPrefix?: string,
): void => {
  if (validationError && validationError.constraints) {
    const constraintsKeys = Object.keys(validationError.constraints);
    for (let i = 0; i < constraintsKeys.length; i += 1) {
      if (parsedErrors.indexOf(validationError.constraints[constraintsKeys[i]]) === -1) {
        parsedErrors.push((parentPrefix || '') + validationError.constraints[constraintsKeys[i]]);
      }
    }
  }
  if (validationError.children && validationError.children.length) {
    const childrenKeys: string[] = Object.keys(validationError.children);
    for (let i = 0; i < childrenKeys.length; i += 1) {
      parseValidationError(
        parsedErrors,
        validationError.children[i],
        generatePrefix(validationError.property, parentPrefix),
      );
    }
  }
};

/**
 * Parse validation errors array
 *
 * @param validationErrors - Array of validation errors
 */
export const parseValidationErrors = (validationErrors: ValidationError[]): void => {
  const parsedErrors: string[] = [];
  for (let i = 0; i < validationErrors.length; i += 1) {
    parseValidationError(parsedErrors, validationErrors[i]);
  }
  throw new NestJSMicroserviceValidationException(parsedErrors);
};
