import { TestController } from './controllers/ test.controller';
import { TargetObject } from '../src/common/types';
import { NestJSMicroserviceValidationException } from '../src';

describe('dto validation', () => {
  let controller: TestController;
  let dto: TargetObject;

  beforeEach(() => {
    dto = {
      nestedField: {
        requiredField: 'string',
      },
      nestedArray: [
        {
          requiredField: 'string',
        },
      ],
      requiredField: 'string',
      optionalField: null,
      isBaseDto: true,
    };

    controller = new TestController();
  });

  it('should pass validation using base dto in case if validation is correct', async () => {
    const result = await controller.createOne({ dto });
    expect(result).toEqual(dto);
  });

  it('should not pass validation in case if dto properties don\'t pass "class-validator" validation', async () => {
    dto.requiredField = 1000;
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string');
  });

  it('should not pass validation in case if dto contains extra fields', async () => {
    dto.extraField = 'string';
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property extraField should not exist');
  });

  it('should not pass validation in case if dto contains fields that are no described by decorators', async () => {
    dto.invisibleField = 'string';
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property invisibleField should not exist');
  });

  it('should not pass validation in case if dto contains properties related to different crud group', async () => {
    dto.groupCheck = true;
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property groupCheck should not exist');
  });

  it('should pass bulk validation correct for createMany method validation in case of correct params', async () => {
    const bulkDto = {
      bulk: [dto],
    };
    const result = await controller.createMany({ dto: bulkDto });
    expect(result).toEqual(bulkDto);
  });

  it('should pass bulk validation correct for createMany method validation in case of correct params', async () => {
    dto.requiredField = 100;
    const bulkDto = {
      bulk: [dto],
    };

    let err: any;
    try {
      await controller.createMany({ dto: bulkDto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('bulk -> 0 -> requiredField must be a string');
  });

  it('should validate crud methods by custom dto in case if its added by argument decorator', async () => {
    let err: any;
    dto.nestedField.nestedGroupRelationCheck = 'string';
    try {
      await controller.updateOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual(
      'property nestedArray should not exist, property optionalField should not exist, property isBaseDto should not exist',
    );
  });

  it('should validate nested dto with the same method group as parent', async () => {
    let err: any;
    try {
      await controller.updateOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual(
      'property nestedArray should not exist, property optionalField should not exist, property isBaseDto should not exist, nestedField -> nestedGroupRelationCheck must be a string',
    );
  });

  it('should not pass validation in case if NESTED dto properties don\'t pass "class-validator" validation', async () => {
    dto.nestedField = {};
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> requiredField must be a string');
  });

  it('should not pass validation in case if NESTED dto contains fields that are no described by decorators', async () => {
    dto.nestedField.extraField = 'string';
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> property extraField should not exist');
  });

  it('should not pass validation in case if NESTED dto has wrong value', async () => {
    dto.nestedField = null;
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nested property nestedField must be either object or array');
  });

  it('should not pass validation in case if NESTED dto array item properties don\'t pass "class-validator" validation', async () => {
    dto.nestedArray[0] = {};
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedArray -> 0 -> requiredField must be a string');
  });

  it('should not pass validation in case if NESTED dto array item contains fields that are no described by decorators', async () => {
    dto.nestedArray[0].extraField = 'string';
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedArray -> 0 -> property extraField should not exist');
  });

  it('should not pass validation in case if second nested dto is added but is disabled by RelatedDto decorator', async () => {
    dto = {
      requiredField: 'string',
      nestedField: {
        requiredField: 'string',
        nestedField: {
          requiredField: 'string',
        },
      },
    };
    let err: any;
    try {
      await controller.withoutRelations({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> property nestedField should not exist');
  });

  it('should pass validation in case if second nested dto is added and enabled by RelatedDto decorator', async () => {
    dto = {
      requiredField: 'string',
      nestedField: {
        requiredField: 'string',
        nestedField: {
          requiredField: 'string',
        },
      },
    };
    const result = await controller.withRelations({ dto });
    expect(result).toEqual(dto);
  });

  it('should pass validation in case if third nested dto is added and dot separated string is added to RelatedDto decorator', async () => {
    dto = {
      requiredField: 'string',
      nestedField: {
        requiredField: 'string',
        nestedField: {
          requiredField: 'string',
          nestedField: {
            requiredField: 'string',
          },
        },
      },
    };
    const result = await controller.withRelations({ dto });
    expect(result).toEqual(dto);
  });

  it('should not pass validation in case if second nested dto has validation issues and nested validation is enabled by RelatedDto decorator', async () => {
    dto = {
      requiredField: 'string',
      nestedField: {
        requiredField: 'string',
        nestedField: {
          requiredField: 'string',
          extraField: 'string',
        },
      },
    };
    let err: any;
    try {
      await controller.withRelations({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual(
      'nestedField -> nestedField -> property extraField should not exist',
    );
  });

  it('should validate array property correct in case if they are empty', async () => {
    dto.nestedArray = [];
    const result = await controller.createOne({ dto });
    expect(result).toEqual(dto);
  });

  it('should validate array property correct in case if string is passed instead of array', async () => {
    dto.nestedArray = 'someString';
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual(
      'each value in nested property nestedArray must be either object or array',
    );
  });
});
