import { TestController } from './controllers/ test.controller';
import { ParsedRequestParams } from '../src/interfaces/parsed-request-params.interface';
import { NestJSMicroserviceValidationException } from '../src';
import { QueryFilterOperators } from '../src/common/constants';

interface Request {
  parsed: ParsedRequestParams;
  options: Record<string, any>;
}

describe('request validation', () => {
  let controller: TestController;
  let req: Request;

  beforeEach(() => {
    req = {
      parsed: {
        fields: [],
        paramsFilter: [],
        search: {},
        filter: [],
        or: [],
        join: [],
        sort: [],
      },
      options: {},
    };

    controller = new TestController();
  });

  it('should pass validation for empty request correctly, without any issues', async () => {
    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should pass validation for request with fields list in case if fields are correct', async () => {
    req.parsed.fields = ['nestedArray', 'nestedField', 'requiredField'];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation in case if field property does not exists in validation dto', async () => {
    req.parsed.fields = ['nestedArray', 'nestedField', 'requiredField', 'extraField'];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property extraField should not exist');
  });

  it('should pass request with nested fields list in case if fields are correct', async () => {
    req.parsed.fields = ['nestedField.requiredField', 'requiredField'];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass request validation with nested fields list in case if field is not exist in validation dto', async () => {
    req.parsed.fields = ['nestedField.extraField', 'requiredField'];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> property extraField should not exist');
  });

  it('should validate request correct if its passed as method argument (not as data property)', async () => {
    req.parsed.fields = ['nestedField.extraField', 'requiredField'];
    let err: any;
    try {
      await controller.getMany(req);
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> property extraField should not exist');
  });

  it('should pass validation for request with sort fields list in case if fields are correct', async () => {
    req.parsed.sort = [{ field: 'requiredField', order: 'ASC' }];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation in case if sort field property does not exists in validation dto', async () => {
    req.parsed.sort = [{ field: 'extraField', order: 'ASC' }];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property extraField should not exist');
  });

  it('should pass request with nested sort fields list in case if fields are correct', async () => {
    req.parsed.sort = [{ field: 'nestedField.requiredField', order: 'ASC' }];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass request validation with nested sort fields list in case if field not exist in validation dto', async () => {
    req.parsed.sort = [{ field: 'nestedField.extraField', order: 'ASC' }];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> property extraField should not exist');
  });

  it('should pass validation for request with join fields list in case if fields contains @RelatedDto decorator', async () => {
    req.parsed.join = [{ field: 'nestedField' }];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation in case if join field property does not exists in validation dto', async () => {
    req.parsed.join = [{ field: 'extraField' }];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property extraField should not exist');
  });

  it('should not pass validation for request with join fields list in case if fields does not contain @RelatedDto decorator', async () => {
    req.parsed.join = [{ field: 'requiredField' }];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('property requiredField is not allowed for join');
  });

  it(
    'should pass validation for request with nested join fields list ' +
      'in case if fields contains @RelatedDto decorator and nested relations are allowed',
    async () => {
      req.parsed.join = [{ field: 'nestedField.nestedField' }];
      const result = await controller.requestWithRelations({ req });
      expect(result).toEqual(req.parsed);
    },
  );
  it(
    'should pass validation for request with nested join fields list ' +
      'in case if fields contains @RelatedDto decorator and second nested relations are allowed',
    async () => {
      req.parsed.join = [{ field: 'nestedField.nestedField.nestedField' }];
      const result = await controller.requestWithRelations({ req });
      expect(result).toEqual(req.parsed);
    },
  );

  it(
    'should not pass validation for request with nested join fields list ' +
      'in case if fields contains @RelatedDto decorator but nested relations are  not allowed',
    async () => {
      req.parsed.join = [{ field: 'nestedField.nestedField' }];
      let err: any;
      try {
        await controller.requestWithoutRelations({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('nestedField -> property nestedField should not exist');
    },
  );

  it(
    'should not pass validation for request with nested join fields list ' +
      'in case if nested fields does not contain @RelatedDto decorator',
    async () => {
      req.parsed.join = [{ field: 'nestedField.requiredField' }];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('nestedField -> property requiredField is not allowed for join');
    },
  );

  it(
    'should not pass validation for request join field if field does not contain @RelatedDto decorator ' +
      'even if field is added to "fields" or "sort" arrays of request',
    async () => {
      req.parsed.join = [{ field: 'nestedField.requiredField' }];
      req.parsed.fields = ['nestedField.requiredField'];
      req.parsed.sort = [{ field: 'nestedField.requiredField', order: 'DESC' }];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('nestedField -> property requiredField is not allowed for join');
    },
  );

  it('should pass validation for param field in case if its nested and correct', async () => {
    req.parsed.filter = [
      { field: 'nestedField.requiredField', operator: QueryFilterOperators.EQ, value: 'string' },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation for param field in case if its nested and don\'t pass "class-validator" validation', async () => {
    req.parsed.filter = [
      { field: 'nestedField.requiredField', operator: QueryFilterOperators.EQ, value: 100 },
    ];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> requiredField must be a string');
  });

  it('should not pass validation for param field in case if its nested and contains property that does not exist in nested validation dto', async () => {
    req.parsed.filter = [
      { field: 'nestedField.extraField', operator: QueryFilterOperators.EQ, value: 'string' },
    ];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> property extraField should not exist');
  });

  it(
    'should pass validation for request with nested param fields list ' +
      'in case if fields contains @RelatedDto decorator and nested relations are allowed',
    async () => {
      req.parsed.filter = [
        {
          field: 'nestedField.nestedField.requiredField',
          operator: QueryFilterOperators.EQ,
          value: 'string',
        },
      ];
      const result = await controller.requestWithRelations({ req });
      expect(result).toEqual(req.parsed);
    },
  );

  it(
    'should not pass validation for request with nested param fields list ' +
      'in case if fields contains @RelatedDto decorator but nested relations are not allowed',
    async () => {
      req.parsed.filter = [
        {
          field: 'nestedField.nestedField.requiredField',
          operator: QueryFilterOperators.EQ,
          value: 'string',
        },
      ];
      let err: any;
      try {
        await controller.requestWithoutRelations({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('nestedField -> property nestedField should not exist');
    },
  );

  it('should pass validation for request param field if field is correct and operator requires array as value', async () => {
    req.parsed.filter = [
      {
        field: 'requiredField',
        operator: QueryFilterOperators.IN,
        value: ['string', 'another string'],
      },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it(
    'should not pass validation for request param field ' +
      'in case if filter properties don\'t pass "class-validator" validation and operator requires array as value',
    async () => {
      req.parsed.filter = [
        {
          field: 'requiredField',
          operator: QueryFilterOperators.IN,
          value: ['string', 100],
        },
      ];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('requiredField must be a string');
    },
  );

  it('should pass validation for request filter param if field is optional and operator does not requires value', async () => {
    req.parsed.filter = [
      {
        field: 'optionalField',
        operator: QueryFilterOperators.IS_NULL,
      },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  // FIELD

  it('should pass validation for optional param field in case if value exists', async () => {
    req.parsed.filter = [
      {
        field: 'optionalField',
        operator: QueryFilterOperators.EQ,
        value: 'string',
      },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation for request param field if field is not optional and operator does not requires value', async () => {
    req.parsed.filter = [
      {
        field: 'requiredField',
        operator: QueryFilterOperators.IS_NULL,
      },
    ];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string');
  });

  it('should pass validation for request filter field if field is correct', async () => {
    req.parsed.filter = [
      { field: 'requiredField', operator: QueryFilterOperators.EQ, value: 'string' },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation for request filter field in case if filter properties don\'t pass "class-validator" validation', async () => {
    req.parsed.filter = [{ field: 'requiredField', operator: QueryFilterOperators.EQ, value: 100 }];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string');
  });

  it(
    'should not pass request validation for request filter field ' +
      'in case if filter contains some fields that do not exist in validation dto',
    async () => {
      req.parsed.filter = [
        { field: 'extraField', operator: QueryFilterOperators.EQ, value: 'string' },
      ];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property extraField should not exist');
    },
  );

  it('should pass validation for request paramsFilter field if field is correct', async () => {
    req.parsed.paramsFilter = [
      { field: 'requiredField', operator: QueryFilterOperators.EQ, value: 'string' },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation for request paramsFilter field in case if paramsFilter properties don\'t pass "class-validator" validation', async () => {
    req.parsed.paramsFilter = [
      { field: 'requiredField', operator: QueryFilterOperators.EQ, value: 100 },
    ];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string');
  });

  it(
    'should not pass request validation for request paramsFilter field ' +
      'in case if paramsFilter contains some fields that do not exist in validation dto',
    async () => {
      req.parsed.paramsFilter = [
        { field: 'extraField', operator: QueryFilterOperators.EQ, value: 'string' },
      ];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property extraField should not exist');
    },
  );

  it('should pass validation for request "or" field if field is correct', async () => {
    req.parsed.or = [
      { field: 'requiredField', operator: QueryFilterOperators.EQ, value: 'string' },
    ];

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation for request "or" field in case if "or" properties don\'t pass "class-validator" validation', async () => {
    req.parsed.or = [{ field: 'requiredField', operator: QueryFilterOperators.EQ, value: 100 }];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string');
  });

  it(
    'should not pass request validation for request "or" field ' +
      'in case if "or" contains some fields that do not exist in validation dto',
    async () => {
      req.parsed.or = [{ field: 'extraField', operator: QueryFilterOperators.EQ, value: 'string' }];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property extraField should not exist');
    },
  );

  it('should not pass request validation if param field value is object', async () => {
    req.parsed.or = [
      {
        field: 'nestedField.requiredField',
        operator: QueryFilterOperators.EQ,
        value: { requiredField: 'string' },
      },
    ];
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual(
      'value for property nestedField.requiredField should be primitive',
    );
  });

  it('should pass validation for request paramsFilter field if field is correct', async () => {
    req.parsed.search = { $and: [{ requiredField: 'string' }] };

    const result = await controller.getOne({ req });

    expect(result).toEqual(req.parsed);
  });

  it('should not pass validation for request paramsFilter field in case if paramsFilter properties don\'t pass "class-validator" validation', async () => {
    req.parsed.search = { $and: [{ requiredField: 100 }] };
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string');
  });

  it(
    'should not pass request validation for request paramsFilter field ' +
      'in case if paramsFilter contains some fields that do not exist in validation dto',
    async () => {
      req.parsed.search = { $and: [{ extraField: 'string' }] };
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property extraField should not exist');
    },
  );

  it('should validate correct in case of nested search structure', async () => {
    req.parsed.search = {
      $and: [
        {
          $or: [
            {
              $and: [{ requiredField: { $eq: 100 } }, { optionalField: { $in: ['string', 100] } }],
            },
            { requiredField: 'string' },
          ],
        },
        { requiredField: 'string' },
      ],
    };
    let err: any;
    try {
      await controller.getOne({ req });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('requiredField must be a string, optionalField must be a string');
  });
});
