import { ValidateNestedOnlyController } from './controllers/validate-nested-only.controller';
import { TargetObject } from '../src/common/types';
import { ParsedRequestParams } from '../src/interfaces/parsed-request-params.interface';
import { QueryFilterOperators } from '../src/common/constants';
import { NestJSMicroserviceValidationException } from '../src';

interface Request {
  parsed: ParsedRequestParams;
  options: Record<string, any>;
}

describe('dto validation', () => {
  let controller: ValidateNestedOnlyController;
  let dto: TargetObject;
  let req: Request;

  beforeEach(() => {
    req = {
      parsed: {
        fields: [],
        paramsFilter: [],
        search: {},
        filter: [],
        or: [],
        join: [],
        sort: [],
      },
      options: {},
    };

    dto = {
      nestedField: {
        optionalField: 'string',
        requiredField: 'string',
      },
      optionalField: 'string',
    };

    controller = new ValidateNestedOnlyController();
  });

  it('should pass validation for main and nested dto in case if request validation is correct', async () => {
    req.parsed.filter = [
      { field: 'requiredField', operator: QueryFilterOperators.EQ, value: 'string' },
      { field: 'nestedField.requiredField', operator: QueryFilterOperators.EQ, value: 'string' },
    ];
    const result = await controller.getOne({ req });
    expect(result).toEqual(req.parsed);
  });

  it(
    'should validate main dto for request in case if group in @ValidateIfRelatedOnly decorator does not' +
      ' match method group',
    async () => {
      req.parsed.filter = [
        { field: 'requiredField', operator: QueryFilterOperators.EQ, value: 100 },
      ];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('requiredField must be a string');
    },
  );

  it(
    'should not pass validation for main dto for request in case if group in @ValidateIfRelatedOnly' +
      ' decorator does not match method group, but pass it for nested dto',
    async () => {
      req.parsed.filter = [
        { field: 'optionalField', operator: QueryFilterOperators.EQ, value: 'string' },
        { field: 'nestedField.optionalField', operator: QueryFilterOperators.EQ, value: 'string' },
      ];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property optionalField should not exist');
    },
  );

  it('should pass validation for main and nested dto in case if request fields validation is correct', async () => {
    req.parsed.fields = ['requiredField', 'nestedField.requiredField'];
    const result = await controller.getOne({ req });
    expect(result).toEqual(req.parsed);
  });

  it(
    'should not pass validation for main dto for request fields in case if group in @ValidateIfRelatedOnly' +
      ' decorator does not match method group, but pass it for nested dto',
    async () => {
      req.parsed.fields = ['optionalField', 'nestedField.optionalField'];
      let err: any;
      try {
        await controller.getOne({ req });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property optionalField should not exist');
    },
  );

  it('should pass dto validation for main and nested dto in case if validation is correct', async () => {
    const result = await controller.createOne({ dto });
    expect(result).toEqual(dto);
  });

  it(
    'should not allow property in dto validation for main dto in case if group in @ValidateIfRelatedOnly' +
      ' decorator match method group',
    async () => {
      dto.requiredField = 'string';
      let err: any;
      try {
        await controller.createOne({ dto });
      } catch (e) {
        err = e;
      }
      expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
      expect(err?.message).toEqual('property requiredField should not exist');
    },
  );

  it('should validate property correctly in case if its nested dto and @ValidateIfRelatedOnly and request group matches', async () => {
    dto.nestedField.requiredField = 100;
    let err: any;
    try {
      await controller.createOne({ dto });
    } catch (e) {
      err = e;
    }
    expect(err).toBeInstanceOf(NestJSMicroserviceValidationException);
    expect(err?.message).toEqual('nestedField -> requiredField must be a string');
  });
});
